__doc__   = """Photo object"""
__version = '$ Release 0.0.3 $'[-7:-2]

############################################################################### 
#
# Copyright (c) 1999 Andrew Lahser
# All rights reserved. Written by Andrew Lahser andrew@apl-software.com
# 
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions 
# are met: 
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer. 
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution. 
# 3. The name of the author may not be used to endorse or promote products 
#    derived from this software without specific prior written permission 
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
#   
#  In accordance with the license provided for by the software upon 
#  which some of the source code has been derived or used, the following 
#  acknowledgement is hereby provided : 
# 
#      "This product includes software developed by Digital Creations 
#      for use in the Z Object Publishing Environment 
#      (http://www.zope.org/)." 
#  
############################################################################### 
   

from OFS.content_types import guess_content_type
from OFS.Image import File, Image, cookId, Pdata
from Globals import HTMLFile, MessageDialog
from OFS.PropertyManager import PropertyManager
from AccessControl.Role import RoleManager
from webdav.common import rfc1123_date
from OFS.SimpleItem import Item_w__name__
from cStringIO import StringIO
from Globals import Persistent
from Acquisition import Implicit
from DateTime import DateTime
from types import IntType, StringType, TupleType
                                                        
import Globals, string, struct, OFS.content_types
import PIL

StringType=type('')

manage_addPhotoForm=HTMLFile('PhotoAdd',globals(),Kind='Photo',kind='Photo')

def manage_addPhoto(self, id, file, title='', displays=None, precondition='', 
                    content_type='', REQUEST=None):
    """
    Add a new Photo object.

    Creates a new Photo object 'id' with the contents of 'file'.
    """
    id, title = cookId(id, title, file)

    self=self.this()

    # First, we create the image without data:
    self._setObject(id, Photo(id,title,'',displays,content_type, precondition))
        
    # Now we "upload" the data.  By doing this in two steps, we
    # can use a database trick to make the upload more efficient.
    self._getOb(id).manage_upload(file)
    
    if REQUEST is not None:
        try:    url=self.DestinationURL()
        except: url=REQUEST['URL1']
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
    return id

 
class Photo(Image):
    """Photo objects are images that use PIL to resize themselves for a variety of 
       display sizes.
    """
    meta_type='Photo'
    _properties=({'id':'title', 'type': 'string'},
                 {'id':'content_type', 'type':'string'},
                 {'id':'height', 'type':'string'},
                 {'id':'width', 'type':'string'},
                 )
    
    manage_options=({'label':'Edit', 'action':'manage_main'},
                    {'label':'Upload', 'action':'manage_uploadForm'},
                    {'label':'Displays', 'action':'manage_displays'},
                    {'label':'Properties', 'action':'manage_propertiesForm'},
                    {'label':'View', 'action':'view_image_or_file'},
                    {'label':'Security', 'action':'manage_access'},
                    )

    manage_displays=HTMLFile('displays',globals())

    def __init__(self, id, title, file, displays = None, content_type = '', precondition = ''):
        self.__name__=id
        self.title=title
        self.precondition=precondition
       
        data, size = self._read_data(file)
        content_type=self._get_content_type(file, data, id, content_type)
	if displays == None:
            self._displays={'thumbnail':(128,128), 'xlarge': (1263, 822), 'small': (783, 398), 
	              'webtv': (544, 378), 'medium': (1007, 566), 'xsmall': (623, 278), 
		      'large': (1135, 668)}
	else:
            self._displays=displays
        self.update_data(data, content_type, size)
 
    def update_data(self, data, content_type = None, size = None):
        """
        Replaces the current contents of the Photo object with file.
        The file or images contents are replaced with the contents of 'file'.
        """
        Image.update_data(self, data, content_type, size)
        # Recreate all of the display sizes for a new image
        for display in self._displays.keys():
            (xsize, ysize) = self._displays[display]
	    self.editDisplay(display, xsize, ysize)
    
    def addDisplay(self, display, xsize, ysize):
        """ Add a display for this photo """
        if self._displays.has_key(display):
            raise 'Bad Request', 'Invalid or duplicate display name'
        self.editDisplay(display, xsize, ysize)
 
    def editDisplay(self, display, xsize, ysize):
        """ Add a display for this photo """
        setattr(self, 'data_' + display, self.resize(xsize, ysize))
        self._displays[display] = (xsize, ysize)

    def delDisplay(self, display):
        """ Delete a display from this photo """
        delattr(self, 'data_' + display)
	displays = self._displays
	del displays[display]
	self._displays = displays
       
    def displayMap(self):
        """Return a tuple of mappings"""
	displayMap = []
        for display in self._displays.keys():
	    (xsize, ysize) = self._displays[display]
            displayMap = displayMap +  [{ 'id' : display, 'xsize' : xsize, 'ysize' : ysize}]
	displayMap.sort(lambda a,b: cmp(a['xsize'],b['xsize']))
	return tuple(displayMap)
    
    def resize(self, xsize = 125, ysize = 125, REQUEST = None):
        """Return a scaled image"""
        if (self.size <= 0):
            return None
        # Unwrap PData if size bigger than 1<<16
	if type(self.data) == StringType:
            im = PIL.Image.open(StringIO(self.data))
	else:
            r = (self.data.data)
            next = self.data.next
            while next is not None:
	        self=next
                r.append(self.data)
		next=self.next
            r = string.join(r,'')
            im = PIL.Image.open(StringIO(r))
	# Preserve Aspect Ratio
	(actual_xsize, actual_ysize) = im.size
        if ( xsize / actual_xsize ) <  ( ysize / actual_ysize ):
            ysize = actual_ysize * xsize / actual_xsize
        else:
            xsize = actual_xsize * ysize / actual_ysize
	# Resize image using PIL
	fmt = im.format
        im = im.resize((xsize,ysize))
        outFile = StringIO()
        im.save(outFile, fmt)
        outFile.seek(0)
        outString = outFile.read()
	# Set Response Headers if HTTP request
        if (REQUEST is not None):
            RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
            RESPONSE.setHeader('Content-Type', self.content_type)
        return outString

#   Web Interface
    def manage_editDisplay(self, REQUEST = None):
        """ Edit displays """
        for display in self._displays.keys():
	    xsize = REQUEST.get('xsize-' + display)
	    ysize = REQUEST.get('ysize-' + display)
	    (actual_xsize, actual_ysize) = self._displays[display]
            self.editDisplay(display, xsize, ysize)
        return MessageDialog(
               title  ='Success!',
               message='Your changes have been saved',
               action ='manage_displays')
	
    def manage_addDisplay(self, id, xsize, ysize, REQUEST = None):
        """ Add a display """
	display = REQUEST.get('id')
        if display is None:
	    raise 'BadRequest', 'The display name does not exist.'
	xsize = REQUEST.get('xsize')
        xsize = int(xsize)
	if xsize is None:
	    raise 'BadRequest', 'The display width does not exist.'
	ysize = REQUEST.get('ysize')
        ysize = int(ysize)
        if display is None:
	    raise 'BadRequest', 'The display height does not exist.'
        self.addDisplay(display, xsize, ysize)
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_displays')
	
    def manage_delDisplay(self, ids, REQUEST = None):
        """ Delete one or more displays by id """
        if ids is None:
            return MessageDialog(
                   title='No displays specified',
                   message='No diplsays were selected for deletion!',
                   action ='./manage_displays',)
        for id in ids:
            if not self._displays.has_key(id):
                raise 'BadRequest', (
                      'The display <em>%s</em> does not exist' % id)
            if id == "thumbnail":
                raise 'BadRequest', (
                      'The display <em>%s</em> may not be deleted' % id)
	    self.delDisplay(id)

        if REQUEST is not None:
            return self.manage_displays(self, REQUEST)

    def index_html(self, REQUEST, RESPONSE, display = None ):
        """
        The default view of the contents of the Photo

        Returns the contents of the Photo at the requested display size.
	Also, sets the Content-Type HTTP header to the objects content type.
        """
        if display is not None and self._displays.has_key(display):
            # Attempt to handle If-Modified-Since headers.
            ms=REQUEST.get_header('If-Modified-Since', None)
            if ms is not None:
                ms=string.split(ms, ';')[0]
                ms=DateTime(ms).timeTime()
                if self._p_mtime > ms:
                    RESPONSE.setStatus(304)
                    return RESPONSE
            # Set the response header for the photo
            RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
            RESPONSE.setHeader('Content-Type', self.content_type)
	    return getattr(self, 'data_' + display)
        else:
	    # Inherit the Image behavior
            return Image.index_html(self, REQUEST, RESPONSE)

    def thumbnail(self, REQUEST, RESPONSE):
        """Return a photo thumbnail"""
        RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
        RESPONSE.setHeader('Content-Type', self.content_type)
        return self.data_thumbnail

